'''
rules.py

A library of functions for enforcing and enacting the documentation rules.

THE RULES
You cannot write to a CirrusVolume unless you've passed it:
(1) A set of sources from which this was created (e.g., another CloudVolume
path or a freeform justification like "tracer annotation"). This must be
formatted as a list[str], and any sources that haven't been logged previously
will be added to the current sources field of the provenance file

(2) The motivation for creating the volume (str) has been logged and you've
included that motivation when instantiating the object. A volume can have
multiple motivation notes, and your motivation only need match one of them.

(3) You've passed a Process (a code environment & parameters) to the class.
The process will be logged unless another process with the same task
description has already been logged. These are defined by provenance-tools.
'''
from __future__ import annotations

import warnings
from typing import Optional

import cloudvolume as cv
import provenancetoolbox as ptb


def check_writing_rules(sources: Optional[list[str]] = None,
                        motivation: Optional[str] = None,
                        process: Optional[ptb.Process] = None
                        ) -> None:
    '''
    Checks whether the provided fields are sufficient to allow writing
    to a volume.
    '''
    assert all(v is not None for v in [sources, motivation, process]
               ), (""
                   "Need to define sources, motivation and process in order"
                   " to write to this volume")

    # Checking sources
    assert isinstance(sources, list)
    assert all(isinstance(v, str) for v in sources)

    # Checking motivation
    assert isinstance(motivation, str)

    # Checking process
    assert isinstance(process, ptb.Process)


def add_sources(cloudvolume: cv.CloudVolume,
                sources: Optional[list[str]] = None
                ) -> None:
    '''Documents the sources of a CloudVolume if they don't already exist'''
    currentsources = set(cloudvolume.provenance.sources)

    newsources = currentsources.union(sources)

    if len(newsources) > len(currentsources):
        cloudvolume.provenance.sources = list(newsources)
        cloudvolume.commit_provenance()


def add_motivation(cloudvolume: cv.CloudVolume,
                   motivation: Optional[str] = None
                   ) -> None:
    '''Documents the motivation of a CloudVolume if it doesn't already exist'''
    if pt.note_absent(cloudvolume, motivation, pt.MOTIVATION):
        pt.add_motivation(cloudvolume, motivation)


def add_process(cloudvolume: cv.CloudVolume,
                process: Optional[ptb.Process] = None
                ) -> None:
    '''
    Adds the current process to the CloudVolume documentation if it doesn't
    already exist
    '''
    if pt.process_absent(cloudvolume, process.description):
        pt.log_process(cloudvolume, process)
    else:
        warnings.warn('Process with the same description already logged.'
                      ' Skipping')


def documentvolume(cloudvolume: cv.CloudVolume,
                   sources: Optional[list[str]] = None,
                   motivation: Optional[str] = None,
                   process: Optional[ptb.Process] = None
                   ) -> None:
    'A single function to perform default documentation of a volume'
    add_sources(cloudvolume, sources)
    add_motivation(cloudvolume, motivation)
    add_process(cloudvolume, process)



'''
volume.py

A wrapper for the generic CloudVolume class. Returns a separate wrapper for
each registered CloudVolume front end.
'''
from __future__ import annotations

from typing import Optional

import cloudvolume as cv
import provenancetoolbox as ptb

from . import rules


REGISTERED_PLUGINS = dict()


def register_plugin(key, creation_function):
    REGISTERED_PLUGINS[key] = creation_function


class CloudVolume(cv.CloudVolume):
    def __new__(cls,
                *args,
                sources: Optional[list[str]] = None,
                motivation: Optional[str] = None,
                process: Optional[ptb.Process] = None,
                **kwargs):
        cloudvolume = cv.CloudVolume(*args, **kwargs)

        return REGISTERED_PLUGINS[type(cloudvolume)](
                   cloudvolume, sources, motivation, process)

    # need to re-define this bc it automatically writes
    @classmethod
    def from_numpy(cls,
                   *args,
                   sources: Optional[List[str]] = None,
                   motivation: Optional[str] = None,
                   process: ptb.Process = None,
                   **kwargs):
        rules.check_writing_rules(sources, motivation, process)

        cloudvolume = super().from_numpy(cls, *args, **kwargs)

        rules.add_sources(cloudvolume, sources)
        rules.add_motivation(cloudvolume, motivation)
        rules.add_process(cloudvolume, process)

        return REGISTERED_PLUGINS[type(cloudvolume)](
                   cloudvolume, sources, motivation, process)



'''
graphene.py

A wrapper for CloudVolumeGraphene
NOT IMPLEMENTED YET
'''
from __future__ import annotations

import warnings
from typing import Optional

import cloudvolume as cv
import provenancetoolbox as ptb
from cloudvolume.frontends.graphene import CloudVolumeGraphene

from . import rules
from .volume import register_plugin



def register():
    register_plugin(CloudVolumeGraphene, CirrusVolumeGraphene)


class CirrusVolumeGraphene(CloudVolumeGraphene):

    def __new__(self,
                cloudvolume: CloudVolume,
                sources: Optional[list[str]] = None,
                motivation: Optional[str] = None,
                process: Optional[ptb.Process] = None
                ):
        warnings.warn("CirrusVolumeGraphene not implemented yet!"
                      " Passing you a normal CloudVolumeGraphene")
        return cloudvolume
